@@ -10,15 +10,16 @@
 #include "include/xlist.h"
 
 #include "messages/MClientCapRelease.h"
+#include "mds/MDSMap.h"
 
 struct Cap;
-class Inode;
+struct Inode;
 struct CapSnap;
 struct MetaRequest;
 class MClientCapRelease;
 
 struct MetaSession {
-  int mds_num;
+  mds_rank_t mds_num;
   ConnectionRef con;
   version_t seq;
   uint64_t cap_gen;
