@@ -1,3 +1,6 @@
+// -*- mode:C++; tab-width:8; c-basic-offset:2; indent-tabs-mode:t -*-
+// vim: ts=8 sw=2 smarttab
+
 #ifndef CEPH_RGW_USER_H
 #define CEPH_RGW_USER_H
 
@@ -59,7 +62,8 @@ extern int rgw_store_user_info(RGWRados *store, RGWUserInfo& info, RGWUserInfo *
  * returns: 0 on success, -ERR# on failure (including nonexistence)
  */
 extern int rgw_get_user_info_by_uid(RGWRados *store, string& user_id, RGWUserInfo& info,
-                                    RGWObjVersionTracker *objv_tracker = NULL, time_t *pmtime = NULL);
+                                    RGWObjVersionTracker *objv_tracker = NULL, time_t *pmtime = NULL,
+                                    rgw_cache_entry_info *cache_info = NULL);
 /**
  * Given an swift username, finds the user info associated with it.
  * returns: 0 on success, -ERR# on failure (including nonexistence)
@@ -129,6 +133,8 @@ struct RGWUserAdminOpState {
   uint32_t max_buckets;
   __u8 suspended;
   __u8 system;
+  __u8 exclusive;
+  __u8 fetch_stats;
   std::string caps;
   RGWObjVersionTracker objv;
   uint32_t op_mask;
@@ -265,6 +271,12 @@ struct RGWUserAdminOpState {
     system = is_system;
     system_specified = true;
   }
+  void set_exclusive(__u8 is_exclusive) {
+    exclusive = is_exclusive;
+  }
+  void set_fetch_stats(__u8 is_fetch_stats) {
+    fetch_stats = is_fetch_stats;
+  }
   void set_user_info(RGWUserInfo& user_info) {
     user_id = user_info.user_id;
     info = user_info;
@@ -402,6 +414,8 @@ struct RGWUserAdminOpState {
     perm_mask = 0;
     suspended = 0;
     system = 0;
+    exclusive = 0;
+    fetch_stats = 0;
     op_mask = 0;
 
     existing_user = false;
